@@ -3,12 +3,12 @@
  * heap.c
  *	  code to create and destroy POSTGRES heap relations
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/catalog/heap.c,v 1.232.2.2 2003/05/12 00:17:33 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/catalog/heap.c,v 1.253 2003/09/25 06:57:57 petere Exp $
  *
  *
  * INTERFACE ROUTINES
@@ -42,6 +42,7 @@
 #include "catalog/pg_inherits.h"
 #include "catalog/pg_statistic.h"
 #include "catalog/pg_type.h"
+#include "commands/tablecmds.h"
 #include "commands/trigger.h"
 #include "miscadmin.h"
 #include "nodes/makefuncs.h"
@@ -146,17 +147,16 @@ static Form_pg_attribute SysAtt[] = {&a1, &a2, &a3, &a4, &a5, &a6, &a7};
 
 /*
  * This function returns a Form_pg_attribute pointer for a system attribute.
- * Note that we elog if the presented attno is invalid.
+ * Note that we elog if the presented attno is invalid, which would only
+ * happen if there's a problem upstream.
  */
 Form_pg_attribute
 SystemAttributeDefinition(AttrNumber attno, bool relhasoids)
 {
 	if (attno >= 0 || attno < -(int) lengthof(SysAtt))
-		elog(ERROR, "SystemAttributeDefinition: invalid attribute number %d",
-			 attno);
+		elog(ERROR, "invalid system attribute number %d", attno);
 	if (attno == ObjectIdAttributeNumber && !relhasoids)
-		elog(ERROR, "SystemAttributeDefinition: invalid attribute number %d",
-			 attno);
+		elog(ERROR, "invalid system attribute number %d", attno);
 	return SysAtt[-attno - 1];
 }
 
@@ -222,9 +222,11 @@ heap_create(const char *relname,
 	if (!allow_system_table_mods &&
 	(IsSystemNamespace(relnamespace) || IsToastNamespace(relnamespace)) &&
 		IsNormalProcessingMode())
-		elog(ERROR, "cannot create %s.%s: "
-			 "system catalog modifications are currently disallowed",
-			 get_namespace_name(relnamespace), relname);
+		ereport(ERROR,
+				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+				 errmsg("permission denied to create \"%s.%s\"",
+						get_namespace_name(relnamespace), relname),
+				 errdetail("System catalog modifications are currently disallowed.")));
 
 	/*
 	 * Real ugly stuff to assign the proper relid in the relation
@@ -337,7 +339,7 @@ heap_storage_create(Relation rel)
  *
  *		this is used to make certain the tuple descriptor contains a
  *		valid set of attribute names and datatypes.  a problem simply
- *		generates elog(ERROR) which aborts the current transaction.
+ *		generates ereport(ERROR) which aborts the current transaction.
  * --------------------------------
  */
 void
@@ -349,8 +351,10 @@ CheckAttributeNamesTypes(TupleDesc tupdesc, char relkind)
 
 	/* Sanity check on column count */
 	if (natts < 0 || natts > MaxHeapAttributeNumber)
-		elog(ERROR, "Number of columns is out of range (0 to %d)",
-			 MaxHeapAttributeNumber);
+		ereport(ERROR,
+				(errcode(ERRCODE_TOO_MANY_COLUMNS),
+				 errmsg("tables can have at most %d columns",
+						MaxHeapAttributeNumber)));
 
 	/*
 	 * first check for collision with system attribute names
@@ -364,8 +368,10 @@ CheckAttributeNamesTypes(TupleDesc tupdesc, char relkind)
 		{
 			if (SystemAttributeByName(NameStr(tupdesc->attrs[i]->attname),
 									  tupdesc->tdhasoid) != NULL)
-				elog(ERROR, "name of column \"%s\" conflicts with an existing system column",
-					 NameStr(tupdesc->attrs[i]->attname));
+				ereport(ERROR,
+						(errcode(ERRCODE_DUPLICATE_COLUMN),
+						 errmsg("column name \"%s\" conflicts with a system column name",
+								NameStr(tupdesc->attrs[i]->attname))));
 		}
 	}
 
@@ -378,8 +384,10 @@ CheckAttributeNamesTypes(TupleDesc tupdesc, char relkind)
 		{
 			if (strcmp(NameStr(tupdesc->attrs[j]->attname),
 					   NameStr(tupdesc->attrs[i]->attname)) == 0)
-				elog(ERROR, "column name \"%s\" is duplicated",
-					 NameStr(tupdesc->attrs[j]->attname));
+				ereport(ERROR,
+						(errcode(ERRCODE_DUPLICATE_COLUMN),
+						 errmsg("column name \"%s\" is duplicated",
+								NameStr(tupdesc->attrs[j]->attname))));
 		}
 	}
 
@@ -410,25 +418,34 @@ CheckAttributeType(const char *attname, Oid atttypid)
 	 * Warn user, but don't fail, if column to be created has UNKNOWN type
 	 * (usually as a result of a 'retrieve into' - jolly)
 	 *
-	 * Refuse any attempt to create a pseudo-type column or one that uses
-	 * a standalone composite type.  (Eventually we should probably refuse
+	 * Refuse any attempt to create a pseudo-type column or one that uses a
+	 * standalone composite type.  (Eventually we should probably refuse
 	 * all references to complex types, but for now there's still some
 	 * Berkeley-derived code that thinks it can do this...)
 	 */
 	if (atttypid == UNKNOWNOID)
-		elog(WARNING, "Attribute \"%s\" has an unknown type"
-			 "\n\tProceeding with relation creation anyway",
-			 attname);
+		ereport(WARNING,
+				(errcode(ERRCODE_INVALID_TABLE_DEFINITION),
+				 errmsg("column \"%s\" has type \"unknown\"", attname),
+				 errdetail("Proceeding with relation creation anyway.")));
 	else if (att_typtype == 'p')
-		elog(ERROR, "Attribute \"%s\" has pseudo-type %s",
-			 attname, format_type_be(atttypid));
+	{
+		/* Special hack for pg_statistic: allow ANYARRAY during initdb */
+		if (atttypid != ANYARRAYOID || IsUnderPostmaster)
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_TABLE_DEFINITION),
+					 errmsg("column \"%s\" has pseudo-type %s",
+							attname, format_type_be(atttypid))));
+	}
 	else if (att_typtype == 'c')
 	{
-		Oid		typrelid = get_typ_typrelid(atttypid);
+		Oid			typrelid = get_typ_typrelid(atttypid);
 
 		if (get_rel_relkind(typrelid) == RELKIND_COMPOSITE_TYPE)
-			elog(ERROR, "Attribute \"%s\" has composite type %s",
-				 attname, format_type_be(atttypid));
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_TABLE_DEFINITION),
+					 errmsg("column \"%s\" has composite type %s",
+							attname, format_type_be(atttypid))));
 	}
 }
 
@@ -670,6 +687,8 @@ AddNewRelationType(const char *typeName,
 			   ',',				/* default array delimiter */
 			   F_RECORD_IN,		/* input procedure */
 			   F_RECORD_OUT,	/* output procedure */
+			   F_RECORD_RECV,	/* receive procedure */
+			   F_RECORD_SEND,	/* send procedure */
 			   InvalidOid,		/* array element type - irrelevant */
 			   InvalidOid,		/* domain base type - irrelevant */
 			   NULL,			/* default type value - none */
@@ -694,6 +713,7 @@ heap_create_with_catalog(const char *relname,
 						 TupleDesc tupdesc,
 						 char relkind,
 						 bool shared_relation,
+						 OnCommitAction oncommit,
 						 bool allow_system_table_mods)
 {
 	Relation	pg_class_desc;
@@ -709,7 +729,9 @@ heap_create_with_catalog(const char *relname,
 	CheckAttributeNamesTypes(tupdesc, relkind);
 
 	if (get_relname_relid(relname, relnamespace))
-		elog(ERROR, "Relation '%s' already exists", relname);
+		ereport(ERROR,
+				(errcode(ERRCODE_DUPLICATE_TABLE),
+				 errmsg("relation \"%s\" already exists", relname)));
 
 	/*
 	 * Create the relcache entry (mostly dummy at this point) and the
@@ -795,6 +817,12 @@ heap_create_with_catalog(const char *relname,
 	 */
 	StoreConstraints(new_rel_desc, tupdesc);
 
+	/*
+	 * If there's a special on-commit action, remember it
+	 */
+	if (oncommit != ONCOMMIT_NOOP)
+		register_on_commit_action(new_rel_oid, oncommit);
+
 	/*
 	 * ok, the relation has been cataloged, so close our relations and
 	 * return the oid of the newly created relation.
@@ -860,8 +888,7 @@ DeleteRelationTuple(Oid relid)
 						 ObjectIdGetDatum(relid),
 						 0, 0, 0);
 	if (!HeapTupleIsValid(tup))
-		elog(ERROR, "DeleteRelationTuple: cache lookup failed for relation %u",
-			 relid);
+		elog(ERROR, "cache lookup failed for relation %u", relid);
 
 	/* delete the relation tuple from pg_class, and finish up */
 	simple_heap_delete(pg_class_desc, &tup->t_self);
@@ -940,7 +967,7 @@ RemoveAttributeById(Oid relid, AttrNumber attnum)
 							   Int16GetDatum(attnum),
 							   0, 0);
 	if (!HeapTupleIsValid(tuple))		/* shouldn't happen */
-		elog(ERROR, "RemoveAttributeById: Failed to find attribute %d in relation %u",
+		elog(ERROR, "cache lookup failed for attribute %d of relation %u",
 			 attnum, relid);
 	attStruct = (Form_pg_attribute) GETSTRUCT(tuple);
 
@@ -948,12 +975,13 @@ RemoveAttributeById(Oid relid, AttrNumber attnum)
 	attStruct->attisdropped = true;
 
 	/*
-	 * Set the type OID to invalid.  A dropped attribute's type link cannot
-	 * be relied on (once the attribute is dropped, the type might be too).
-	 * Fortunately we do not need the type row --- the only really essential
-	 * information is the type's typlen and typalign, which are preserved in
-	 * the attribute's attlen and attalign.  We set atttypid to zero here
-	 * as a means of catching code that incorrectly expects it to be valid.
+	 * Set the type OID to invalid.  A dropped attribute's type link
+	 * cannot be relied on (once the attribute is dropped, the type might
+	 * be too). Fortunately we do not need the type row --- the only
+	 * really essential information is the type's typlen and typalign,
+	 * which are preserved in the attribute's attlen and attalign.  We set
+	 * atttypid to zero here as a means of catching code that incorrectly
+	 * expects it to be valid.
 	 */
 	attStruct->atttypid = InvalidOid;
 
@@ -964,7 +992,8 @@ RemoveAttributeById(Oid relid, AttrNumber attnum)
 	attStruct->attstattarget = 0;
 
 	/* Change the column name to something that isn't likely to conflict */
-	snprintf(newattname, NAMEDATALEN, "........pg.dropped.%d........", attnum);
+	snprintf(newattname, sizeof(newattname),
+			 "........pg.dropped.%d........", attnum);
 	namestrcpy(&(attStruct->attname), newattname);
 
 	simple_heap_update(attr_rel, &tuple->t_self, tuple);
@@ -1031,7 +1060,7 @@ RemoveAttrDefault(Oid relid, AttrNumber attnum,
 	heap_close(attrdef_rel, RowExclusiveLock);
 
 	if (complain && !found)
-		elog(ERROR, "RemoveAttrDefault: no default found for rel %u attnum %d",
+		elog(ERROR, "could not find attrdef tuple for relation %u attnum %d",
 			 relid, attnum);
 }
 
@@ -1067,8 +1096,7 @@ RemoveAttrDefaultById(Oid attrdefId)
 
 	tuple = systable_getnext(scan);
 	if (!HeapTupleIsValid(tuple))
-		elog(ERROR, "RemoveAttrDefaultById: cache lookup failed for attrdef %u",
-			 attrdefId);
+		elog(ERROR, "could not find tuple for attrdef %u", attrdefId);
 
 	myrelid = ((Form_pg_attrdef) GETSTRUCT(tuple))->adrelid;
 	myattnum = ((Form_pg_attrdef) GETSTRUCT(tuple))->adnum;
@@ -1090,8 +1118,8 @@ RemoveAttrDefaultById(Oid attrdefId)
 							   Int16GetDatum(myattnum),
 							   0, 0);
 	if (!HeapTupleIsValid(tuple))		/* shouldn't happen */
-		elog(ERROR, "RemoveAttrDefaultById: cache lookup failed for rel %u attr %d",
-			 myrelid, myattnum);
+		elog(ERROR, "cache lookup failed for attribute %d of relation %u",
+			 myattnum, myrelid);
 
 	((Form_pg_attribute) GETSTRUCT(tuple))->atthasdef = false;
 
@@ -1145,8 +1173,7 @@ heap_drop_with_catalog(Oid rid)
 	 */
 	i = FlushRelationBuffers(rel, (BlockNumber) 0);
 	if (i < 0)
-		elog(ERROR, "heap_drop_with_catalog: FlushRelationBuffers returned %d",
-			 i);
+		elog(ERROR, "FlushRelationBuffers returned %d", i);
 
 	/*
 	 * remove inheritance information
@@ -1168,6 +1195,11 @@ heap_drop_with_catalog(Oid rid)
 	 */
 	DeleteRelationTuple(RelationGetRelid(rel));
 
+	/*
+	 * forget any ON COMMIT action for the rel
+	 */
+	remove_on_commit_action(rid);
+
 	/*
 	 * unlink the relation's physical file and finish up.
 	 */
@@ -1261,7 +1293,7 @@ StoreAttrDefault(Relation rel, AttrNumber attnum, char *adbin)
 								Int16GetDatum(attnum),
 								0, 0);
 	if (!HeapTupleIsValid(atttup))
-		elog(ERROR, "cache lookup of attribute %d in relation %u failed",
+		elog(ERROR, "cache lookup failed for attribute %d of relation %u",
 			 attnum, RelationGetRelid(rel));
 	attStruct = (Form_pg_attribute) GETSTRUCT(atttup);
 	if (!attStruct->atthasdef)
@@ -1370,7 +1402,7 @@ StoreRelCheck(Relation rel, char *ccname, char *ccbin)
 						  ' ',
 						  ' ',
 						  ' ',
-						  InvalidOid, /* no associated index */
+						  InvalidOid,	/* no associated index */
 						  expr, /* Tree form check constraint */
 						  ccbin,	/* Binary form check constraint */
 						  ccsrc);		/* Source form check constraint */
@@ -1515,11 +1547,14 @@ AddRelationRawConstraints(Relation rel,
 
 			ccname = cdef->name;
 			/* Check against pre-existing constraints */
-			if (ConstraintNameIsUsed(RelationGetRelid(rel),
+			if (ConstraintNameIsUsed(CONSTRAINT_RELATION,
+									 RelationGetRelid(rel),
 									 RelationGetNamespace(rel),
 									 ccname))
-				elog(ERROR, "constraint \"%s\" already exists for relation \"%s\"",
-					 ccname, RelationGetRelationName(rel));
+				ereport(ERROR,
+						(errcode(ERRCODE_DUPLICATE_OBJECT),
+						 errmsg("constraint \"%s\" for relation \"%s\" already exists",
+								ccname, RelationGetRelationName(rel))));
 			/* Check against other new constraints */
 			/* Needed because we don't do CommandCounterIncrement in loop */
 			foreach(listptr2, rawConstraints)
@@ -1532,8 +1567,10 @@ AddRelationRawConstraints(Relation rel,
 					cdef2->name == NULL)
 					continue;
 				if (strcmp(cdef2->name, ccname) == 0)
-					elog(ERROR, "Duplicate CHECK constraint name: '%s'",
-						 ccname);
+					ereport(ERROR,
+							(errcode(ERRCODE_DUPLICATE_OBJECT),
+						 errmsg("check constraint \"%s\" already exists",
+								ccname)));
 			}
 		}
 		else
@@ -1549,7 +1586,8 @@ AddRelationRawConstraints(Relation rel,
 				 * pre-existing constraints, nor with any auto-generated
 				 * names so far.
 				 */
-				ccname = GenerateConstraintName(RelationGetRelid(rel),
+				ccname = GenerateConstraintName(CONSTRAINT_RELATION,
+												RelationGetRelid(rel),
 												RelationGetNamespace(rel),
 												&constr_name_ctr);
 
@@ -1585,27 +1623,28 @@ AddRelationRawConstraints(Relation rel,
 		/*
 		 * Make sure it yields a boolean result.
 		 */
-		expr = coerce_to_boolean(expr, "CHECK");
+		expr = coerce_to_boolean(pstate, expr, "CHECK");
 
 		/*
 		 * Make sure no outside relations are referred to.
 		 */
 		if (length(pstate->p_rtable) != 1)
-			elog(ERROR, "Only relation \"%s\" can be referenced in CHECK constraint expression",
-				 relname);
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
+					 errmsg("only table \"%s\" can be referenced in check constraint",
+							relname)));
 
 		/*
 		 * No subplans or aggregates, either...
 		 */
-		if (contain_subplans(expr))
-			elog(ERROR, "cannot use subselect in CHECK constraint expression");
-		if (contain_agg_clause(expr))
-			elog(ERROR, "cannot use aggregate function in CHECK constraint expression");
-
-		/*
-		 * Might as well try to reduce any constant expressions.
-		 */
-		expr = eval_const_expressions(expr);
+		if (pstate->p_hasSubLinks)
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				   errmsg("cannot use subquery in check constraint")));
+		if (pstate->p_hasAggs)
+			ereport(ERROR,
+					(errcode(ERRCODE_GROUPING_ERROR),
+					 errmsg("cannot use aggregate function in check constraint")));
 
 		/*
 		 * Constraints are evaluated with execQual, which expects an
@@ -1615,11 +1654,6 @@ AddRelationRawConstraints(Relation rel,
 		 */
 		expr = (Node *) make_ands_implicit((Expr *) expr);
 
-		/*
-		 * Must fix opids in operator clauses.
-		 */
-		fix_opids(expr);
-
 		/*
 		 * OK, store it.
 		 */
@@ -1660,7 +1694,7 @@ SetRelationNumChecks(Relation rel, int numchecks)
 								ObjectIdGetDatum(RelationGetRelid(rel)),
 								0, 0, 0);
 	if (!HeapTupleIsValid(reltup))
-		elog(ERROR, "cache lookup of relation %u failed",
+		elog(ERROR, "cache lookup failed for relation %u",
 			 RelationGetRelid(rel));
 	relStruct = (Form_pg_class) GETSTRUCT(reltup);
 
@@ -1691,9 +1725,9 @@ SetRelationNumChecks(Relation rel, int numchecks)
  * in the expression.  (Even though we plan to reject vars, it's more
  * user-friendly to give the correct error message than "unknown var".)
  *
- * If atttypid is not InvalidOid, check that the expression is coercible
- * to the specified type.  atttypmod is needed in this case, and attname
- * is used in the error message if any.
+ * If atttypid is not InvalidOid, coerce the expression to the specified
+ * type (and typmod atttypmod).   attname is only needed in this case:
+ * it is used in the error message, if any.
  */
 Node *
 cookDefault(ParseState *pstate,
@@ -1715,60 +1749,55 @@ cookDefault(ParseState *pstate,
 	 * Make sure default expr does not refer to any vars.
 	 */
 	if (contain_var_clause(expr))
-		elog(ERROR, "cannot use column references in DEFAULT clause");
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
+			  errmsg("cannot use column references in default expression")));
 
 	/*
 	 * It can't return a set either.
 	 */
 	if (expression_returns_set(expr))
-		elog(ERROR, "DEFAULT clause must not return a set");
+		ereport(ERROR,
+				(errcode(ERRCODE_DATATYPE_MISMATCH),
+				 errmsg("default expression must not return a set")));
 
 	/*
 	 * No subplans or aggregates, either...
 	 */
-	if (contain_subplans(expr))
-		elog(ERROR, "cannot use subselects in DEFAULT clause");
-	if (contain_agg_clause(expr))
-		elog(ERROR, "cannot use aggregate functions in DEFAULT clause");
+	if (pstate->p_hasSubLinks)
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("cannot use subquery in default expression")));
+	if (pstate->p_hasAggs)
+		ereport(ERROR,
+				(errcode(ERRCODE_GROUPING_ERROR),
+				 errmsg("cannot use aggregate function in default expression")));
 
 	/*
-	 * Check that it will be possible to coerce the expression to the
-	 * column's type.  We store the expression without coercion, however,
-	 * to avoid premature coercion in cases like
-	 *
-	 * CREATE TABLE tbl (fld timestamp DEFAULT 'now'::text);
-	 *
-	 * NB: this should match the code in rewrite/rewriteHandler.c that will
-	 * actually do the coercion, to ensure we don't accept an unusable
-	 * default expression.
+	 * Coerce the expression to the correct type and typmod, if given.
+	 * This should match the parser's processing of non-defaulted
+	 * expressions --- see updateTargetListEntry().
 	 */
 	if (OidIsValid(atttypid))
 	{
 		Oid			type_id = exprType(expr);
 
-		if (coerce_to_target_type(expr, type_id,
-								  atttypid, atttypmod,
-								  COERCION_ASSIGNMENT,
-								  COERCE_IMPLICIT_CAST) == NULL)
-			elog(ERROR, "Column \"%s\" is of type %s"
-				 " but default expression is of type %s"
-				 "\n\tYou will need to rewrite or cast the expression",
-				 attname,
-				 format_type_be(atttypid),
-				 format_type_be(type_id));
+		expr = coerce_to_target_type(pstate, expr, type_id,
+									 atttypid, atttypmod,
+									 COERCION_ASSIGNMENT,
+									 COERCE_IMPLICIT_CAST);
+		if (expr == NULL)
+			ereport(ERROR,
+					(errcode(ERRCODE_DATATYPE_MISMATCH),
+					 errmsg("column \"%s\" is of type %s"
+							" but default expression is of type %s",
+							attname,
+							format_type_be(atttypid),
+							format_type_be(type_id)),
+			errhint("You will need to rewrite or cast the expression.")));
 	}
 
-	/*
-	 * Might as well try to reduce any constant expressions.
-	 */
-	expr = eval_const_expressions(expr);
-
-	/*
-	 * Must fix opids, in case any operators remain...
-	 */
-	fix_opids(expr);
-
-	return (expr);
+	return expr;
 }
 
 
@@ -1883,41 +1912,21 @@ RemoveStatistics(Relation rel, AttrNumber attnum)
 static void
 RelationTruncateIndexes(Oid heapId)
 {
-	Relation	indexRelation;
-	ScanKeyData entry;
-	SysScanDesc scan;
-	HeapTuple	indexTuple;
+	Relation	heapRelation;
+	List	   *indlist;
 
-	/* Scan pg_index to find indexes on specified heap */
-	indexRelation = heap_openr(IndexRelationName, AccessShareLock);
-	ScanKeyEntryInitialize(&entry, 0,
-						   Anum_pg_index_indrelid,
-						   F_OIDEQ,
-						   ObjectIdGetDatum(heapId));
-	scan = systable_beginscan(indexRelation, IndexIndrelidIndex, true,
-							  SnapshotNow, 1, &entry);
+	/*
+	 * Open the heap rel.  We need grab no lock because we assume
+	 * heap_truncate is holding an exclusive lock on the heap rel.
+	 */
+	heapRelation = heap_open(heapId, NoLock);
 
-	while (HeapTupleIsValid(indexTuple = systable_getnext(scan)))
+	/* Ask the relcache to produce a list of the indexes of the rel */
+	foreach(indlist, RelationGetIndexList(heapRelation))
 	{
-		Form_pg_index indexform = (Form_pg_index) GETSTRUCT(indexTuple);
-		Oid			indexId;
+		Oid			indexId = lfirsto(indlist);
+		Relation	currentIndex;
 		IndexInfo  *indexInfo;
-		Relation	heapRelation,
-					currentIndex;
-
-		/*
-		 * For each index, fetch info needed for index_build
-		 */
-		indexId = indexform->indexrelid;
-		indexInfo = BuildIndexInfo(indexform);
-
-		/*
-		 * We have to re-open the heap rel each time through this loop
-		 * because index_build will close it again.  We need grab no lock,
-		 * however, because we assume heap_truncate is holding an
-		 * exclusive lock on the heap rel.
-		 */
-		heapRelation = heap_open(heapId, NoLock);
 
 		/* Open the index relation */
 		currentIndex = index_open(indexId);
@@ -1925,6 +1934,9 @@ RelationTruncateIndexes(Oid heapId)
 		/* Obtain exclusive lock on it, just to be sure */
 		LockRelation(currentIndex, AccessExclusiveLock);
 
+		/* Fetch info needed for index_build */
+		indexInfo = BuildIndexInfo(currentIndex);
+
 		/*
 		 * Drop any buffers associated with this index. If they're dirty,
 		 * they're just dropped without bothering to flush to disk.
@@ -1941,30 +1953,37 @@ RelationTruncateIndexes(Oid heapId)
 
 		/*
 		 * index_build will close both the heap and index relations (but
-		 * not give up the locks we hold on them).
+		 * not give up the locks we hold on them).	We're done with this
+		 * index, but we must re-open the heap rel.
 		 */
+		heapRelation = heap_open(heapId, NoLock);
 	}
 
-	/* Complete the scan and close pg_index */
-	systable_endscan(scan);
-	heap_close(indexRelation, AccessShareLock);
+	/* Finish by closing the heap rel again */
+	heap_close(heapRelation, NoLock);
 }
 
 /*
  *	 heap_truncate
  *
- *	 This routine is used to truncate the data from the
- *	 storage manager of any data within the relation handed
- *	 to this routine.
+ *	 This routine deletes all data within the specified relation.
+ *
+ * This is not transaction-safe!  There is another, transaction-safe
+ * implementation in commands/cluster.c.  We now use this only for
+ * ON COMMIT truncation of temporary tables, where it doesn't matter.
  */
 void
 heap_truncate(Oid rid)
 {
 	Relation	rel;
+	Oid			toastrelid;
 
 	/* Open relation for processing, and grab exclusive access on it. */
 	rel = heap_open(rid, AccessExclusiveLock);
 
+	/* Don't allow truncate on tables that are referenced by foreign keys */
+	heap_truncate_check_FKs(rel);
+
 	/*
 	 * Release any buffers associated with this relation.  If they're
 	 * dirty, they're just dropped without bothering to flush to disk.
@@ -1979,8 +1998,71 @@ heap_truncate(Oid rid)
 	/* If this relation has indexes, truncate the indexes too */
 	RelationTruncateIndexes(rid);
 
+	/* If it has a toast table, recursively truncate that too */
+	toastrelid = rel->rd_rel->reltoastrelid;
+	if (OidIsValid(toastrelid))
+		heap_truncate(toastrelid);
+
 	/*
 	 * Close the relation, but keep exclusive lock on it until commit.
 	 */
 	heap_close(rel, NoLock);
 }
+
+/*
+ * heap_truncate_check_FKs
+ *		Check for foreign keys referencing a relation that's to be truncated
+ *
+ * We disallow such FKs (except self-referential ones) since the whole point
+ * of TRUNCATE is to not scan the individual rows to be thrown away.
+ *
+ * This is split out so it can be shared by both implementations of truncate.
+ * Caller should already hold a suitable lock on the relation.
+ */
+void
+heap_truncate_check_FKs(Relation rel)
+{
+	Oid			relid = RelationGetRelid(rel);
+	ScanKeyData key;
+	Relation	fkeyRel;
+	SysScanDesc fkeyScan;
+	HeapTuple	tuple;
+
+	/*
+	 * Fast path: if the relation has no triggers, it surely has no FKs
+	 * either.
+	 */
+	if (rel->rd_rel->reltriggers == 0)
+		return;
+
+	/*
+	 * Otherwise, must scan pg_constraint.  Right now, this is a seqscan
+	 * because there is no available index on confrelid.
+	 */
+	fkeyRel = heap_openr(ConstraintRelationName, AccessShareLock);
+
+	ScanKeyEntryInitialize(&key, 0,
+						   Anum_pg_constraint_confrelid,
+						   F_OIDEQ,
+						   ObjectIdGetDatum(relid));
+
+	fkeyScan = systable_beginscan(fkeyRel, NULL, false,
+								  SnapshotNow, 1, &key);
+
+	while (HeapTupleIsValid(tuple = systable_getnext(fkeyScan)))
+	{
+		Form_pg_constraint con = (Form_pg_constraint) GETSTRUCT(tuple);
+
+		if (con->contype == CONSTRAINT_FOREIGN && con->conrelid != relid)
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("cannot truncate a table referenced in a foreign key constraint"),
+					 errdetail("Table \"%s\" references \"%s\" via foreign key constraint \"%s\".",
+							   get_rel_name(con->conrelid),
+							   RelationGetRelationName(rel),
+							   NameStr(con->conname))));
+	}
+
+	systable_endscan(fkeyScan);
+	heap_close(fkeyRel, AccessShareLock);
+}
