@@ -7,7 +7,7 @@
 # Author:  Peter Eisentraut <peter_e@gmx.net> 
 # Public domain
 
-# $Header: /cvsroot/pgsql-server/src/bin/pg_config/Attic/pg_config.sh,v 1.6 2001/09/30 22:17:51 momjian Exp $
+# $Header: /cvsroot/pgsql-server/src/bin/pg_config/Attic/pg_config.sh,v 1.8 2003/07/23 08:47:25 petere Exp $
 
 me=`basename $0`
 
@@ -23,9 +23,10 @@ val_version='@version@'
 help="\
 $me provides information about the installed version of PostgreSQL.
 
-Usage: $me --bindir | --includedir | --includedir-server | --libdir | --pkglibdir | --configure | --version
+Usage:
+  $me OPTION...
 
-Operation modes:
+Options:
   --bindir              show location of user executables
   --includedir          show location of C header files of the client
                         interfaces
@@ -34,12 +35,13 @@ Operation modes:
   --pkglibdir           show location of dynamically loadable modules
   --configure           show options given to 'configure' script when
                         PostgreSQL was built
-  --version             show the PostgreSQL version and exit
+  --version             show the PostgreSQL version, then exit
+  --help                show this help, then exit
 
 Report bugs to <pgsql-bugs@postgresql.org>."
 
 advice="\
-Try '$me --help' for more information."
+Try \"$me --help\" for more information."
 
 if test "$#" -eq 0 ; then
     echo "$me: argument required" 1>&2
