@@ -6,7 +6,7 @@
  * Copyright (c) 2003, PostgreSQL Global Development Group
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/interfaces/jdbc/org/postgresql/fastpath/Fastpath.java,v 1.16.2.2 2003/12/18 03:29:12 davec Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/interfaces/jdbc/org/postgresql/fastpath/Attic/Fastpath.java,v 1.16.2.4 2004/08/11 06:56:00 jurka Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -118,8 +118,10 @@ public class Fastpath
 				switch (c)
 				{
 					case 'A':	// Asynchronous Notify
-						int pid = stream.ReceiveInteger(4);
+						int msglen = stream.ReceiveIntegerR(4);
+						int pid = stream.ReceiveIntegerR(4);
 						String msg = stream.ReceiveString(conn.getEncoding());
+						String param = stream.ReceiveString(conn.getEncoding());
 						conn.addNotification(new org.postgresql.core.Notification(msg, pid));
 						break;
 						//------------------------------
@@ -140,7 +142,8 @@ public class Fastpath
 						// Notice from backend
 					case 'N':
 						int l_nlen = stream.ReceiveIntegerR(4);
-						conn.addWarning(conn.getEncoding().decode(stream.Receive(l_nlen-4)));
+						PSQLException notify = PSQLException.parseServerError(conn.getEncoding().decode(stream.Receive(l_nlen-4)));
+						conn.addWarning(notify.getMessage());
 						break;
 
 					case 'V':
@@ -233,8 +236,9 @@ public class Fastpath
 				{
 					case 'A':	// Asynchronous Notify
 						//TODO: do something with this
-						int pid = stream.ReceiveInteger(4);
+						int pid = stream.ReceiveIntegerR(4);
 						String msg = stream.ReceiveString(conn.getEncoding());
+						conn.addNotification(new org.postgresql.core.Notification(msg, pid));
 						break;
 
 						//------------------------------
