@@ -9,7 +9,7 @@
  * Copyright (c) 2003, PostgreSQL Global Development Group
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/interfaces/jdbc/org/postgresql/jdbc2/AbstractJdbc2ResultSet.java,v 1.25.2.2 2003/12/18 03:35:55 davec Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/interfaces/jdbc/org/postgresql/jdbc2/Attic/AbstractJdbc2ResultSet.java,v 1.25.2.8 2004/06/21 03:11:49 jurka Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -18,6 +18,8 @@ package org.postgresql.jdbc2;
 import java.io.CharArrayReader;
 import java.io.InputStream;
 import java.io.IOException;
+import java.io.InputStreamReader;
+import java.io.UnsupportedEncodingException;
 import java.math.BigDecimal;
 import java.sql.*;
 import java.util.Enumeration;
@@ -52,9 +54,9 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 	private PreparedStatement selectStatement = null;
 
   
-	public AbstractJdbc2ResultSet(BaseStatement statement, Field[] fields, Vector tuples, String status, int updateCount, long insertOID, boolean binaryCursor)
+	public AbstractJdbc2ResultSet(BaseStatement statement, Field[] fields, Vector tuples, String status, int updateCount, long insertOID)
 	{
-		super (statement, fields, tuples, status, updateCount, insertOID, binaryCursor);
+		super (statement, fields, tuples, status, updateCount, insertOID);
 	}
 
 	public java.net.URL getURL(int columnIndex) throws SQLException
@@ -210,6 +212,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 		rowBuffer = new byte[this_row.length][];
 		System.arraycopy(this_row, 0, rowBuffer, 0, this_row.length);
+		onInsertRow = false;
 
 		return true;
 	}
@@ -220,6 +223,10 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		final int rows_size = rows.size();
 		if (rows_size > 0)
 			current_row = rows_size;
+
+		onInsertRow = false;
+		this_row = null;
+		rowBuffer = null;
 	}
 
 
@@ -227,6 +234,10 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 	{
 		if (rows.size() > 0)
 			current_row = -1;
+
+		onInsertRow = false;
+		this_row = null;
+		rowBuffer = null;
 	}
 
 
@@ -235,12 +246,12 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		if (rows.size() <= 0)
 			return false;
 
-		onInsertRow = false;
 		current_row = 0;
 		this_row = (byte[][]) rows.elementAt(current_row);
 
 		rowBuffer = new byte[this_row.length][];
 		System.arraycopy(this_row, 0, rowBuffer, 0, this_row.length);
+		onInsertRow = false;
 
 		return true;
 	}
@@ -483,6 +494,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 		rowBuffer = new byte[this_row.length][];
 		System.arraycopy(this_row, 0, rowBuffer, 0, this_row.length);
+		onInsertRow = false;
 
 		return true;
 	}
@@ -490,8 +502,14 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 	public boolean previous() throws SQLException
 	{
-		if (--current_row < 0)
+		if (current_row-1 < 0) {
+			current_row = -1;
+			this_row = null;
+			rowBuffer = null;
 			return false;
+		} else {
+			current_row--;
+		}
 		this_row = (byte[][]) rows.elementAt(current_row);
 		rowBuffer = new byte[this_row.length][];
 		System.arraycopy(this_row, 0, rowBuffer, 0, this_row.length);
@@ -580,6 +598,8 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		deleteStatement.executeUpdate();
 
 		rows.removeElementAt(current_row);
+		current_row--;
+		moveToCurrentRow();
 	}
 
 
@@ -734,22 +754,19 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 
 	public boolean rowDeleted() throws SQLException
 	{
-		// only sub-classes implement CONCURuPDATEABLE
-		throw Driver.notImplemented();
+		return false;
 	}
 
 
 	public boolean rowInserted() throws SQLException
 	{
-		// only sub-classes implement CONCURuPDATEABLE
-		throw Driver.notImplemented();
+		return false;
 	}
 
 
 	public boolean rowUpdated() throws SQLException
 	{
-		// only sub-classes implement CONCURuPDATEABLE
-		throw Driver.notImplemented();
+		return false;
 	}
 
 
@@ -759,21 +776,39 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 											  )
 	throws SQLException
 	{
-		byte[] theData = null;
+		if (x == null)
+		{
+			updateNull(columnIndex);
+			return;
+		}
+
 		try
 		{
-			x.read(theData, 0, length);
+			InputStreamReader reader = new InputStreamReader(x, "ASCII");
+			char data[] = new char[length];
+			int numRead = 0;
+			while (true)
+			{
+				int n = reader.read(data, numRead, length - numRead);
+				if (n == -1)
+					break;
+
+				numRead += n;
+
+				if (numRead == length)
+					break;
+			}
+			updateString(columnIndex, new String(data, 0, numRead));
 		}
-		catch (NullPointerException ex )
+		catch (UnsupportedEncodingException uee)
 		{
-			throw new PSQLException("postgresql.updateable.inputstream");
+			throw new PSQLException("postgresql.unusual", PSQLState.UNEXPECTED_ERROR, uee);
 		}
 		catch (IOException ie)
 		{
 			throw new PSQLException("postgresql.updateable.ioerror", ie);
 		}
 
-		updateValue(columnIndex, theData);
 	}
 
 
@@ -791,21 +826,46 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 											   )
 	throws SQLException
 	{
-		byte[] theData = null;
+		if (x == null)
+		{
+			updateNull(columnIndex);
+			return;
+		}
+
+		byte data[] = new byte[length];
+		int numRead = 0;
 		try
 		{
-			x.read(theData, 0, length);
+			while (true)
+			{
+				int n = x.read(data, numRead, length - numRead);
+				if (n == -1)
+					break;
 
-		}
-		catch ( NullPointerException ex )
-		{
-			throw new PSQLException("postgresql.updateable.inputstream");
+				numRead += n;
+
+				if (numRead == length)
+					break;
+			}
 		}
 		catch (IOException ie)
 		{
 			throw new PSQLException("postgresql.updateable.ioerror", ie);
 		}
-		updateValue(columnIndex, theData);
+
+		if (numRead == length)
+		{
+			updateBytes(columnIndex, data);
+		}
+		else
+		{
+			// the stream contained less data than they said
+			// perhaps this is an error?
+			byte data2[] = new byte[numRead];
+			System.arraycopy(data, 0, data2, 0, numRead);
+			updateBytes(columnIndex, data2);
+		}
+
 	}
 
 
@@ -838,21 +898,33 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 												  )
 	throws SQLException
 	{
-		char[] theData = null;
+		if (x == null)
+		{
+			updateNull(columnIndex);
+			return;
+		}
+
 		try
 		{
-			x.read(theData, 0, length);
+			char data[] = new char[length];
+			int numRead = 0;
+			while (true)
+			{
+				int n = x.read(data, numRead, length - numRead);
+				if (n == -1)
+					break;
 
-		}
-		catch (NullPointerException ex)
-		{
-			throw new PSQLException("postgresql.updateable.inputstream");
+				numRead += n;
+
+				if (numRead == length)
+					break;
+			}
+			updateString(columnIndex, new String(data, 0, numRead));
 		}
 		catch (IOException ie)
 		{
 			throw new PSQLException("postgresql.updateable.ioerror", ie);
 		}
-		updateValue(columnIndex, theData);
 	}
 
 
@@ -934,73 +1006,68 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		{
 			throw new PSQLException( "postgresql.updateable.notupdateable" );
 		}
+		if (onInsertRow)
+			throw new PSQLException("postgresql.res.oninsertrow");
 
-		try
+		if (isBeforeFirst() || isAfterLast())
+			return;
+
+		StringBuffer selectSQL = new StringBuffer( "select ");
+
+		final int numColumns = java.lang.reflect.Array.getLength(fields);
+
+		for (int i = 0; i < numColumns; i++ )
 		{
-			StringBuffer selectSQL = new StringBuffer( "select ");
 
-			final int numColumns = java.lang.reflect.Array.getLength(fields);
+			selectSQL.append( fields[i].getName() );
 
-			for (int i = 0; i < numColumns; i++ )
+			if ( i < numColumns - 1 )
 			{
 
-				selectSQL.append( fields[i].getName() );
-
-				if ( i < numColumns - 1 )
-				{
-
-					selectSQL.append(", ");
-				}
-
+				selectSQL.append(", ");
 			}
-			selectSQL.append(" from " ).append(tableName).append(" where ");
-
-			int numKeys = primaryKeys.size();
-
-			for ( int i = 0; i < numKeys; i++ )
-			{
-
-				PrimaryKey primaryKey = ((PrimaryKey) primaryKeys.get(i));
-				selectSQL.append(primaryKey.name).append("= ?");
-
-				if ( i < numKeys - 1 )
-				{
-					selectSQL.append(" and ");
-				}
-			}
-			if ( Driver.logDebug )
-				Driver.debug("selecting " + selectSQL.toString());
-			selectStatement = ((java.sql.Connection) connection).prepareStatement(selectSQL.toString());
-
-
-			for ( int j = 0, i = 1; j < numKeys; j++, i++)
-			{
-				selectStatement.setObject( i, ((PrimaryKey) primaryKeys.get(j)).getValue() );
-			}
-
-			AbstractJdbc2ResultSet rs = (AbstractJdbc2ResultSet) selectStatement.executeQuery();
-
-			if ( rs.first() )
-			{
-				rowBuffer = rs.rowBuffer;
-			}
-
-			rows.setElementAt( rowBuffer, current_row );
-			this_row = rowBuffer;
-			if ( Driver.logDebug )
-				Driver.debug("done updates");
-
-			rs.close();
-			selectStatement.close();
-			selectStatement = null;
 
 		}
-		catch (Exception e)
+		selectSQL.append(" from " ).append(tableName).append(" where ");
+
+		int numKeys = primaryKeys.size();
+
+		for ( int i = 0; i < numKeys; i++ )
 		{
-			if ( Driver.logDebug )
-				Driver.debug(e.getClass().getName() + e);
-			throw new SQLException( e.getMessage() );
+
+			PrimaryKey primaryKey = ((PrimaryKey) primaryKeys.get(i));
+			selectSQL.append(primaryKey.name).append("= ?");
+
+			if ( i < numKeys - 1 )
+			{
+				selectSQL.append(" and ");
+			}
 		}
+		if ( Driver.logDebug )
+			Driver.debug("selecting " + selectSQL.toString());
+		selectStatement = ((java.sql.Connection) connection).prepareStatement(selectSQL.toString());
+
+
+		for ( int j = 0, i = 1; j < numKeys; j++, i++)
+		{
+			selectStatement.setObject( i, ((PrimaryKey) primaryKeys.get(j)).getValue() );
+		}
+
+		AbstractJdbc2ResultSet rs = (AbstractJdbc2ResultSet) selectStatement.executeQuery();
+
+		if ( rs.first() )
+		{
+			rowBuffer = rs.rowBuffer;
+		}
+
+		rows.setElementAt( rowBuffer, current_row );
+		this_row = rowBuffer;
+		if ( Driver.logDebug )
+			Driver.debug("done updates");
+
+		rs.close();
+		selectStatement.close();
+		selectStatement = null;
 
 	}
 
@@ -1012,6 +1079,10 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		{
 			throw new PSQLException( "postgresql.updateable.notupdateable" );
 		}
+		if (isBeforeFirst() || isAfterLast())
+		{
+			throw new PSQLException("postgresql.updateable.badupdateposition");
+		}
 
 		if (doingUpdates)
 		{
@@ -1093,11 +1164,11 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 				updateValues.clear();
 				doingUpdates = false;
 			}
-			catch (Exception e)
+			catch (SQLException e)
 			{
 				if ( Driver.logDebug )
 					Driver.debug(e.getClass().getName() + e);
-				throw new SQLException( e.getMessage() );
+				throw e;
 			}
 
 		}
@@ -1493,6 +1564,7 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 					case Types.NUMERIC:
 					case Types.REAL:
 					case Types.TINYINT:
+					case Types.OTHER:
 
 						rowBuffer[columnIndex] = connection.getEncoding().encode(String.valueOf( valueObject));
 
@@ -1518,6 +1590,10 @@ public abstract class AbstractJdbc2ResultSet extends org.postgresql.jdbc1.Abstra
 		{
 			throw new PSQLException( "postgresql.updateable.notupdateable" );
 		}
+		if (!onInsertRow && (isBeforeFirst() || isAfterLast()))
+		{
+			throw new PSQLException("postgresql.updateable.badupdateposition");
+		}
 		doingUpdates = !onInsertRow;
 		if (value == null)
 			updateNull(columnIndex);
