@@ -3,12 +3,12 @@
  * fe-lobj.c
  *	  Front-end large object interface
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpq/fe-lobj.c,v 1.41 2002/06/20 20:29:54 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpq/fe-lobj.c,v 1.44.4.1 2004/03/05 01:54:13 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -396,9 +396,11 @@ lo_import(PGconn *conn, const char *filename)
 	fd = open(filename, O_RDONLY | PG_BINARY, 0666);
 	if (fd < 0)
 	{							/* error */
+		char		sebuf[256];
+
 		printfPQExpBuffer(&conn->errorMessage,
 					   libpq_gettext("could not open file \"%s\": %s\n"),
-						  filename, strerror(errno));
+					  filename, pqStrerror(errno, sebuf, sizeof(sebuf)));
 		return InvalidOid;
 	}
 
@@ -479,9 +481,11 @@ lo_export(PGconn *conn, Oid lobjId, const char *filename)
 	fd = open(filename, O_CREAT | O_WRONLY | O_TRUNC | PG_BINARY, 0666);
 	if (fd < 0)
 	{							/* error */
+		char		sebuf[256];
+
 		printfPQExpBuffer(&conn->errorMessage,
 					   libpq_gettext("could not open file \"%s\": %s\n"),
-						  filename, strerror(errno));
+					  filename, pqStrerror(errno, sebuf, sizeof(sebuf)));
 		(void) lo_close(conn, lobj);
 		return -1;
 	}
@@ -523,6 +527,7 @@ lo_initialize(PGconn *conn)
 	PGresult   *res;
 	PGlobjfuncs *lobjfuncs;
 	int			n;
+	const char *query;
 	const char *fname;
 	Oid			foid;
 
@@ -539,18 +544,35 @@ lo_initialize(PGconn *conn)
 	MemSet((char *) lobjfuncs, 0, sizeof(PGlobjfuncs));
 
 	/*
-	 * Execute the query to get all the functions at once
+	 * Execute the query to get all the functions at once.  In 7.3 and later
+	 * we need to be schema-safe.
 	 */
-	res = PQexec(conn, "select proname, oid from pg_proc	\
-			where proname = 'lo_open'	\
-		   or proname = 'lo_close'	\
-		   or proname = 'lo_creat'	\
-		   or proname = 'lo_unlink'	\
-		   or proname = 'lo_lseek'	\
-		   or proname = 'lo_tell'	\
-		   or proname = 'loread'	\
-		   or proname = 'lowrite'");
-	if (res == (PGresult *) NULL)
+	if (conn->sversion >= 70300)
+		query = "select proname, oid from pg_catalog.pg_proc "
+			"where proname in ("
+			"'lo_open', "
+			"'lo_close', "
+			"'lo_creat', "
+			"'lo_unlink', "
+			"'lo_lseek', "
+			"'lo_tell', "
+			"'loread', "
+			"'lowrite') "
+			"and pronamespace = (select oid from pg_catalog.pg_namespace "
+			"where nspname = 'pg_catalog')";
+	else
+		query = "select proname, oid from pg_proc "
+			"where proname = 'lo_open' "
+			"or proname = 'lo_close' "
+			"or proname = 'lo_creat' "
+			"or proname = 'lo_unlink' "
+			"or proname = 'lo_lseek' "
+			"or proname = 'lo_tell' "
+			"or proname = 'loread' "
+			"or proname = 'lowrite'";
+
+	res = PQexec(conn, query);
+	if (res == NULL)
 	{
 		free(lobjfuncs);
 		return -1;
